#!/bin/false
# UT2003/2004 ServerInfo module in Python
# Copyright 2005 Michael Farrell
#
# Derived from the PHP version and Documentation by Michiel Hendriks
# http://unreal.student.utwente.nl/UT2003-queryspec.html.  Thanks! :-)
#
# Using the library is simple, just call retrieveStatus with the server IP
# address.  The optional second parameter is the query port, which is by
# default 7778.
#
# >>> import ut2004
# >>> ut2004.retrieveStatus('127.0.0.1')
# ({'ServerName': 'lollercoaster', 'GameName': 'Unreal Tournament 2004 (Retail)', 'CurrentPlayers': 0, 'GameType': 'xDeathMatch', 'GameMap': 'DM-1on1-Albatross', 'MaxPlayers': 6, 'GamePort': 7777}, {'MaxSpectators': '2', 'GameStats': 'False', 'AdminName': 'your mother', 'ServerMode': 'dedicated', 'MinPlayers': '0', 'Translocator': 'False', 'ServerVersion': '3355', 'ForceRespawn': 'False', 'EndTimeDelay': '4.00', 'TimeLimit': '10', 'Mutator3': 'MutVehiclePickups', 'Mutator2': 'MutSlomoDeath', 'GamePassword': 'True', 'Mutator0': 'MutBigHead', 'Mutator1': 'MutQuadJump', 'GoalScore': '20', 'AdminEmail': 'abuse@hotmail.com', 'WeaponStay': 'True'})

import string, struct, time
from twisted.internet.protocol import DatagramProtocol
from twisted.internet import reactor

def retrieveStatus(host, qport = 7778, lag = 1):
	class ut2004conn(DatagramProtocol):
		def startProtocol(self):
			# init store
			global storedata
			global waitcount
			waitcount = 0
			storedata = []

			# connect and request info
			self.transport.connect(host, qport)
			self.transport.write(struct.pack("bxxxx", 0x78))
			self.transport.write(struct.pack("bxxxb", 0x78, 0x03))
	        
		def datagramReceived(self, data, (host, port)):
			global storedata

			# collect data
			storedata.append(data)
			if len(storedata) >= 3:
				# we are expecting three packets, stop at that.
				reactor.stop()

		def connectionRefused(self):
			print "Noone listening"

	# 0 means any port, we don't care
	reactor.listenUDP(0, ut2004conn())
	reactor.callLater(lag, reactor.stop)
	reactor.run()

	return parseData(storedata)

def parseData(data):
	#setup some array constants.
	versionStrings = {
		'\x80\x00\x00\x00': 'Unreal Tournament 2004 (Retail)',
		'\x7F\x00\x00\x00': 'Unreal Tournament 2004 (Demo)',
		'\x7E\x00\x00\x00': 'Unreal 2 XMP',
		'\x79\x00\x00\x00': 'Unreal Tournament 2003 (Retail)',
		'\x78\x00\x00\x00': 'Unreal Tournament 2003 (Demo)'}

	try:
		version = versionStrings[data[0][:4]]
	except:
		version = 'Unknown Unreal Game'


	try:
		if data[0][4] == '\x00':
			serverData = parseServerData(data[0], version)
	except:
		# can't parse data, probably an error somewhere else getting data.
		# bail out.
		return 0

	if data[0][4] == '\x01':
		gameData = parseGameData(data[0])

	if data[1][4] == '\x00':
		serverData = parseServerData(data[1], version)

	if data[1][4] == '\x01':
		gameData = parseGameData(data[1])

	return (serverData, gameData)
	
def parseServerData(data, version):
	# List of game types.	
	gameTypes = {
		# UT2004
		"xDeathMatch": "Deathmatch",
		"ASGameInfo": "Assult",
		"xBombingRun": "Bombing Run",
		"xCTFGame": "Capture the Flag",
		"xDoubleDom": "Double Domination",
		"InstagibCTF": "Instagib CTF",
		"Invasion": "Invasion",
		"xLastManStandingGame": "Last Man Standing",
		"xMutantGame": "Mutant",
		"ONSOnslaughtGame": "Onslaught",
		"xTeamGame": "Team Deathmatch",
		"xVehicleCTFGame": "Vehicle CTF"
	}

	# Unpack information from data returned...
	gamePort = struct.unpack("i", data[10:14])[0]
	gameNameLen = struct.unpack("b", data[18])[0]
	x = 19 + gameNameLen
	gameName = data[19:x][:-1]
	gameMapLen = struct.unpack("b", data[x])[0]
	x = x + 1
	gameMap = data[x:x+gameMapLen][:-1]
	x = x + gameMapLen
	gameTypeLen = struct.unpack("b", data[x])[0]
	x = x + 1
	gameType = data[x:x+gameTypeLen][:-1]
	x = x + gameTypeLen
	currentPlayers = struct.unpack("i", data[x:x+4])[0]
	x = x + 4
	maxPlayers = struct.unpack("i", data[x:x+4])[0]
	x = x + 4

	serverInfo = {
		'GameName': version,
		'GamePort': gamePort,
		'ServerName': gameName,
		'GameMap': gameMap,
		'GameType': gameType,
		'CurrentPlayers': currentPlayers,
		'MaxPlayers': maxPlayers }

	# clean up game type
	if gameTypes.has_key(gameType):
		serverInfo['GameType'] = gameTypes[gameType]
	serverInfo['type'] = 'ut2004'
	return serverInfo

def parseGameData(data):
	x = 5
	mutators = 0
	gameinfo = {}
	while x < len(data):	
		# key name
		datalength = struct.unpack('b', data[x])[0]
		x = x + 1
		varname = data[x:x+datalength][:-1]
		x = x + datalength
		if varname == 'Mutator':
			varname = varname + str(mutators)
			mutators = mutators + 1

		# key value
		datalength = struct.unpack('b', data[x])[0]
		x = x + 1
		varvalue = data[x:x+datalength][:-1]
		x = x + datalength
		
		# append to dict
		gameinfo[varname] = varvalue
	return gameinfo
